/**
 ** regexlib.c   Randy Sargent
 **
 ** Simple ui to the GNU regex functions
 **/

/* Simple ui to the GNU regex functions
   Copyright (C) 1991 Randy Sargent

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

In other words, you are welcome to use, share and improve this program.
You are forbidden to forbid anyone else to use, share and improve
what you give them.   Help stamp out software-hoarding!  */

#include CONFIG

#ifdef UNIX

#include <regex.h>
#include <table.h>
#include <stringlb.h>
#include <regexlib.h>

/* should probably be the same as in regex.c */
#define BYTEWIDTH 8

Table regex_table_i;

Int regex_init_module(void)
{
    static Int already_initted= 0;
    if (!already_initted) {
	re_set_syntax(RE_NO_BK_PARENS | RE_NO_BK_VBAR);
	already_initted= 1;
	if (table_init(&regex_table_i, sizeof(char*),
		       sizeof(struct re_pattern_buffer), table_strcmp)) return 1;
    }
    return 0;
}
  
static char upcase[0400] = 
  { 000, 001, 002, 003, 004, 005, 006, 007,
    010, 011, 012, 013, 014, 015, 016, 017,
    020, 021, 022, 023, 024, 025, 026, 027,
    030, 031, 032, 033, 034, 035, 036, 037,
    040, 041, 042, 043, 044, 045, 046, 047,
    050, 051, 052, 053, 054, 055, 056, 057,
    060, 061, 062, 063, 064, 065, 066, 067,
    070, 071, 072, 073, 074, 075, 076, 077,
    0100, 0101, 0102, 0103, 0104, 0105, 0106, 0107,
    0110, 0111, 0112, 0113, 0114, 0115, 0116, 0117,
    0120, 0121, 0122, 0123, 0124, 0125, 0126, 0127,
    0130, 0131, 0132, 0133, 0134, 0135, 0136, 0137,
    0140, 0101, 0102, 0103, 0104, 0105, 0106, 0107,
    0110, 0111, 0112, 0113, 0114, 0115, 0116, 0117,
    0120, 0121, 0122, 0123, 0124, 0125, 0126, 0127,
    0130, 0131, 0132, 0173, 0174, 0175, 0176, 0177,
    0200, 0201, 0202, 0203, 0204, 0205, 0206, 0207,
    0210, 0211, 0212, 0213, 0214, 0215, 0216, 0217,
    0220, 0221, 0222, 0223, 0224, 0225, 0226, 0227,
    0230, 0231, 0232, 0233, 0234, 0235, 0236, 0237,
    0240, 0241, 0242, 0243, 0244, 0245, 0246, 0247,
    0250, 0251, 0252, 0253, 0254, 0255, 0256, 0257,
    0260, 0261, 0262, 0263, 0264, 0265, 0266, 0267,
    0270, 0271, 0272, 0273, 0274, 0275, 0276, 0277,
    0300, 0301, 0302, 0303, 0304, 0305, 0306, 0307,
    0310, 0311, 0312, 0313, 0314, 0315, 0316, 0317,
    0320, 0321, 0322, 0323, 0324, 0325, 0326, 0327,
    0330, 0331, 0332, 0333, 0334, 0335, 0336, 0337,
    0340, 0341, 0342, 0343, 0344, 0345, 0346, 0347,
    0350, 0351, 0352, 0353, 0354, 0355, 0356, 0357,
    0360, 0361, 0362, 0363, 0364, 0365, 0366, 0367,
    0370, 0371, 0372, 0373, 0374, 0375, 0376, 0377
  };

struct re_pattern_buffer *regex_lookup_buffer(char *pattern)
{
    struct re_pattern_buffer *ret;

    ret= table_get(&regex_table_i, &pattern);
    if (!ret) {
	struct re_pattern_buffer buf;

	buf.allocated= 40;
	buf.buffer= malloc(buf.allocated);
	buf.fastmap= malloc(1 << BYTEWIDTH);
	buf.translate= upcase;

	re_compile_pattern(pattern, strlen(pattern), &buf);
	re_compile_fastmap(&buf);
	
	table_set(&regex_table_i, &pattern, &buf);
	ret= table_get(&regex_table_i, &pattern);
	
    }
    return ret;
}

  
Int regex_find_i(char *string, char *pattern, ...)
{
    struct re_pattern_buffer *regex_buf;
    struct re_registers regs;
    va_list ap;
    Int i;
    Int ret;
    
    for (i= 0; i< RE_NREGS; i++) regs.start[i]=-1;
    regex_init_module();
    regex_buf= regex_lookup_buffer(pattern);
    ret= re_match(regex_buf, string, strlen(string), 0, &regs);
    if (ret >= 0) {
	va_start(ap, pattern);
	for (i= 1; regs.start[i] != -1 && i < RE_NREGS; i++) {
	    char **receive;
	    receive= va_arg(ap, char**);
	    if (receive) {
		*receive= string_substr(string,
					regs.start[i],
					regs.end[i]);
	    }
	}
    }
    return ret;
}

#endif /* #ifdef UNIX */
